#! usr/bin/python
# -*- coding:utf8 -*-
"""
读取文件的脚本

参考日志行：
[pid: 6814|app: 0|req: 1990/15726]
110.90.29.160 () {60 vars in 2074 bytes}
[Mon Sep  9 10:17:57 2019] POST /api/order/order_detail => generated 2558 bytes in 83 msecs
(HTTP/1.1 200) 7 headers in 403 bytes (2 switches on core 1)
"""

import os
import re
import asyncio

import yagmail


class Filereader:
    """读取文件"""
    def __init__(self, filename, logreader, mailsender, wait_time=1.0):
        """初始化日志阅读器，并检验是否为正确的日志文件
        :param filename: string 文件地址
        :param logreader: Logreader object 日志解析类
        :param wait_time: int default=1 跟踪文件变化的间隔时间
        """
        self.filename = filename
        self.logreader = logreader
        self.mailsender = mailsender
        self.wait_time = wait_time
        self.error_log = []

        self._validate_filename()

    async def tail(self):
        """实现tail命令，跟踪文件动态"""
        with open(self.filename) as f:
            f.seek(0, 2)
            i = 0
            while True:
                current_position = f.tell()
                line = f.readline()
                if not line:
                    f.seek(current_position)
                    await asyncio.sleep(self.wait_time)
                else:
                    i += 1
                    await self.read_log(line)

    async def read_log(self, line):
        """读取日志，根据情况进行处理
        :param line: string logline
        """
        result = self.logreader.get_log_info(line)
        # 未匹配到结果，非正常日志信息，记录
        if result is None:
            self.error_log.append(line)
        # 如果出现状态码500，表示错误返回，此时连同之前记录的错误信息，一同通知收件人
        elif result.get('resp_status', '') == '500':
            # 通知收件人
            self.error_log.append(line)
            self._send_email()
            # 清空错误日志记录
            self.error_log = []

    def _send_email(self):
        """发送邮件通知收件人"""
        self.mailsender.send_email(subject='异常日志通知', contents=''.join(self.error_log))

    def _validate_filename(self):
        """确认文件存在、可读、并且不是文件夹"""
        if not os.access(self.filename, os.F_OK):
            raise FileNotFoundError('日志文件 ' + self.filename + ' 不存在')
        if not os.access(self.filename, os.R_OK):
            raise FileNotFoundError('日志文件 ' + self.filename + ' 不可读')
        if os.path.isdir(self.filename):
            raise FileNotFoundError(self.filename + ' 是文件夹而非日志文件')


class Logreader:
    """分析日志"""

    def __init__(self, re_pattern=r''):
        self.pattern = re.compile(re_pattern, re.VERBOSE)

    def get_log_info(self, line):
        """用正则表达式解析log，如果满足表达式，返回解析结果，否则返回None"""
        result = self.pattern.search(line)
        if result:
            return result.groupdict()
        else:
            return result


class Mailsender:
    """邮件处理类"""
    def __init__(self, emailconfig):
        """初始化邮件类
        :param emailconfig: Emailconfig
        """
        self.yagmail = yagmail.SMTP(emailconfig.sender, emailconfig.password,
                                    host=emailconfig.host, port=emailconfig.port)
        self.recipients = emailconfig.recipients

    def send_email(self, subject, contents):
        """发送邮件
        :param subject: string email subject
        :param contents: string message to send
        """
        self.yagmail.send(to=self.recipients, subject=subject, contents=contents)



#! /usr/bin/python
# -*- coding:utf8 -*-

"""监控脚本"""

import asyncio
import argparse
import configparser

from tail_uwsgi_log.reader import Filereader, Logreader, Mailsender
from tail_uwsgi_log.config import Emailconfig, Logconfig


async def monitor(configs):
    """监控日志文件"""
    # 每个文件都创建对应任务
    tasks = []
    for config in configs:
        emailconfig = Emailconfig(recipients=config.get('mail_recipients', []),
                                  sender=config.get('mail_sender', ''),
                                  password=config.get('mail_password', ''),
                                  host=config.get('mail_host', ''),
                                  port=config.get('mail_port', ''),)
        mailsender = Mailsender(emailconfig=emailconfig)
        logconfig = Logconfig(filepath=config.get('filepath', ''), pattern=config.get('pattern', ''),
                              wait_time=config.get('wait_time', 1.0))

        logreader = Logreader(re_pattern=logconfig.pattern)
        filereader = Filereader(filename=logconfig.filepath, logreader=logreader, mailsender=mailsender,
                                wait_time=logconfig.wait_time)
        tasks.append(filereader.tail())

    # 执行任务
    await asyncio.gather(*tasks)


def tail_uwsgi_log():
    """Script to parse configs and execute log tail"""
    parser = argparse.ArgumentParser(prog='tail_uwsgi_log', description='Tail uwsgi logs')
    parser.add_argument('-c', '--config', type=str, help='Config file path')

    # parse configs
    configs = parse_config(parser.parse_args().config)

    # start tailing tasks
    asyncio.run(monitor(configs))


def parse_config(filepath):
    """从配置文件中读取信息"""
    conf = configparser.ConfigParser()
    conf.read(filepath)

    configs = []
    for section in conf.sections():
        if section.startswith('log'):
            config = {
                'filepath': conf.get(section, 'filepath') if conf.has_option(section, 'filepath') else '',
                'wait_time': conf.getfloat(section, 'wait_time') if conf.has_option(section, 'wait_time') else 1,
                'pattern': conf.get(section, 'pattern') if conf.has_option(section, 'pattern') else '',
            }
            config.update(parse_mail_config(conf, section))
            configs.append(config)

    return configs


def parse_mail_config(conf, section):
    """读取邮件配置信息"""
    mail_info = {
        'mail_host': '',
        'mail_port': '',
        'mail_sender': '',
        'mail_password': '',
        'mail_recipients': '',
    }
    if conf.has_section('mail'):
        for k in mail_info.keys():
            if conf.has_option('mail', k):
                mail_info[k] = conf.get('mail', k)

    for k in mail_info.keys():
        if conf.has_option(section, k):
            mail_info[k] = conf.get(section, k)

    # 整理数据格式
    try:
        mail_info['mail_port'] = int(mail_info['mail_port'])
    except ValueError:
        mail_info['mail_port'] = 465
    mail_info['mail_recipients'] = [i.strip() for i in mail_info.get('mail_recipients', '').split(',')]

    return mail_info



#! /usr/bin/python
# -*- coding:utf8 -*-


class Emailconfig:
    """邮件发送配置类"""
    def __init__(self, recipients, sender='', password='', host='smtp.gmail.com', port=465):
        """
        :param recipients: list 收件人列表
        :param sender: string 发送人
        :param password: string 登录密码
        """
        self.recipients = recipients
        self.sender = sender
        self.password = password
        self.host = host
        self.port = port


class Logconfig:
    """日志文件解析配置类，如文件地址、解析方式、错误通知人等"""
    pattern = r'''\]\ (?P<ip>.*?)\ (.*)\ {.*?}\ \[(?P<datetime>.*?)\]\ (?P<request_method>POST|GET|DELETE|PUT|PATCH|OPTIONS)\s
            (?P<request_uri>[^ ]*?)\ =>\ generated\ (?:.*?)\ in\ (?P<resp_msecs>\d+)\ msecs .*\s
            \(HTTP/[\d.]+\ (?P<resp_status>\d+)\)'''

    def __init__(self, filepath, pattern='', wait_time=1.0):
        self.filepath = filepath
        self.wait_time = wait_time

        if pattern is not None and '' != pattern:
            self.pattern = pattern



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='tail_uwsgi_log',
    version='0.0.4',
    packages=find_packages(),

    install_requires=['yagmail>=0.11.220'],

    author='Kant',
    author_email='kant@kantli.com',

    description='Tail several uwsgi log files and send an email when error occurs.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords='uwsgi log',
    url='https://github.com/kant-li/uwsgi_log_analysis',
    license='MIT License',

    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',

    entry_points={
        "console_scripts": [
            "tail_uwsgi_log = tail_uwsgi_log.monitor:tail_uwsgi_log",
        ]
    },
)
